#ifndef _TRANSKODE_COMMONDEFS_H_
#define _TRANSKODE_COMMONDEFS_H_

/***************************************************************************
 *   Copyright (C) 2005 by Sergio Pistone                                  *
 *   sergio_pistone@yahoo.com.ar                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
	#include <config.h>
#endif

#include <qstring.h>
#include <qstringlist.h>
#include <qmutex.h>


//#define LOG_FILE QDir::homeDirPath() + "/Desktop/transkode.log"

#ifdef LOG_FILE

	#define NEVER		0x00000000
	#define ALWAYS		0xFFFFFFFF
	#define ALMOST_NONE	0x80000000

	#define APP			0x00000001
	#define AMAROK		0x00000002
	#define WORKER		0x00000004
	#define STREAMRIP	0x00000008
	#define ENCODE		0x00000010
	#define DECODE		0x00000020
	#define TRANSCODE	0x00000030

	#define COMMON		0x00000040
	#define DIALOG		0x00000080
	#define CONFIG 		0x00000100

	#define SYSTEM		0x00000200
	//#define TRANSC	0x00000400

	#define CURRENTLY_DEBUGGING (ALWAYS)

	#define LOG( x, y ) if ( CURRENTLY_DEBUGGING & (y) ) writeToLogFile( (x) )
	#define RESET_LOG() clearLogFile()

	void clearLogFile();
	void writeToLogFile( const QString& message );
#else
	#define LOG( x, y )
	#define RESET_LOG()
#endif

namespace GeneralOpt
{
	typedef enum {
		/*base_dir = 0, */naming_scheme = 0,
		notify, confirm, overwrite,
		case_insensitive, replace_chars,
		rip_duration, rip_incomplete,
		proc_priority, show_tray, log,
		SIZE,
		UNKNOWN
	} TGeneralOpt;

	QString toQString( TGeneralOpt option );
	TGeneralOpt fromQString( const QString& option );
};

namespace ProfileOpt
{
	typedef enum { naming_scheme=0, encode_type, extension, switches, show, SIZE, UNKNOWN } TProfileOpt;

	QString toQString( TProfileOpt option );
	TProfileOpt fromQString( const QString& option );
};

namespace EncodeType
{
	// NOTE: Keep in sync with src/configdialog/profilesettings.ui!
	typedef enum { decode=0, bypass, lame, mppenc, oggenc, faac, speexenc, mac, flac, wavpack, ttaenc, gstlaunch, SIZE, UNKNOWN } TEncodeType;

	QString toQString( TEncodeType encodeType );
	TEncodeType fromQString( const QString& encodeType );
};

namespace Program
{
	typedef enum {
		lame=0, flac, mac, mplayer, gstlaunch,
		oggenc, oggdec, speexenc, speexdec,
		mppenc, mppdec, faac, faad, wavpack,
		wvunpack, ttaenc, streamripper,
		SIZE,
		UNKNOWN
	} TProgram;

	QString toQString( TProgram program );
	TProgram fromQString( const QString& program );
};

namespace CaseFunc
{
	// Keep in sync with tagsettings.ui!
	typedef enum { leaveasis=0, titlecase, capitalize, lowercase, uppercase, SIZE, UNKNOWN } TCaseFunc;

	QString toQString( TCaseFunc caseFunction );
	TCaseFunc fromQString( const QString& caseFunction );
};

namespace Tag
{
	typedef enum { artist=0, album, year, track, title, genre, comment, SIZE, UNKNOWN } TTag;

	QString toQString( TTag tag );
	TTag fromQString( const QString& tag );
};


class String
{
	public:
		static QString title( const QString& text );
		static QString capitalize( const QString& text );

		static bool isValidCaseExpression( const QString& expression );
		static QString evaluateCaseExpression( const QString& expression );
		static QString evaluateCase( const QString& text, const QString& caseFunction );
		static QString evaluateCase( const QString& text, CaseFunc::TCaseFunc caseFunction );

		// Result should be interpreted by shell literally and
		// AS ONE ENCLOSED ARGUMENT WHEN (ONLY WHEN) NOT BEING SURROUNDED BY "
		static QString quoteForShell( const QString& text );

		static QString padNumber( const QString& string, unsigned width );

		static QString deepCopy( const QString& string );
		static QStringList deepCopy( const QStringList& stringList );

	protected:
		static int rfindFunctionStart( const QString& text );
		static int rfindFunctionEnd( const QString& text, int startPos );
};

class System
{
	public:
		// FALSE ON ERROR
		static bool copy( const QString& srcPath, const QString& dstPath, bool srcVFAT=false, bool dstVFAT=false );
		static bool move( const QString& srcPath, const QString& dstPath, bool srcVFAT=false, bool dstVFAT=false );
		static bool remove( const QString& path, bool VFAT=false );
		static bool recursiveMakeDir( const QString& path, QStringList* createdDirsList=0, bool VFAT=false );

		static bool isReadable( const QString& path );
		static bool isWritable( const QString& path );

		static QString user();

		// Checks if the given path exists; returns the existing path if it does or else returns ""
		// If VFAT is true, path will be searched in a case insensitive way but the returned value
		// will have the actual case of the existing path.
		static QString conflictivePath( const QString& path, bool VFAT=false, bool potentialFiles=false, const QString& exception=QString::null );

		// If the path exists, appends a suffix before the extension wich is incremented
		// until the resulting path doesn't exists. If onlyDir is true the resulting
		// path will not conflict with a dir but may conflict with a file.
		static QString nonConflictivePath( const QString& path, bool onlyDir, bool VFAT=false, bool potentialFiles=false, const QString& exception=QString::null );

		static void addPotentialFile( const QString& path );
		static void delPotentialFile( const QString& path );
		static void clearPotentialFiles();
		static QString isPotentialFile( const QString& path, bool VFAT=false );

	protected:
		static QString correctVFATPath( const QString& path );

		static QStringList s_potentialFiles;
		static QMutex s_potentialFilesMutex;
};

#endif
