/***************************************************************************
 *   Copyright (C) 2005 by Sergio Pistone                                  *
 *   sergio_pistone@yahoo.com.ar                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "transkodeworker.h"

#include <kapplication.h>

int TransKodeWorker::s_idCount = 0;
QMutex TransKodeWorker::s_idCountMutex;

int TransKodeWorker::getUnivocalID()
{
	s_idCountMutex.lock();

	int id = s_idCount++;

	s_idCountMutex.unlock();

	return id;
}


TransKodeWorker::TransKodeWorker( const TransKodeConfig& cfg ):
	QObject( 0, "TransKodeWorker" ),
	QThread(),
	m_workerID( getUnivocalID() ),
	m_config( cfg ),
	m_threadPaused( 0 ),
	m_threadStopped( false ),
	m_currentJob( 0 )
{
	m_jobsList.setAutoDelete( false );

	connect( &m_config, SIGNAL( removedProfiles(const QStringList&) ), this, SLOT( slotProfilesRemoved(const QStringList&) ) );
	connect( &m_config, SIGNAL( changedProfiles(const QStringList&) ), this, SLOT( slotProfilesModified(const QStringList&) ) );
	connect( &m_config, SIGNAL( removedSuppEncodeTypes(const QValueList<int>&) ), this, SLOT( slotEncTypesRemoved(const QValueList<int>&) ) );
}

TransKodeWorker::~TransKodeWorker()
{
	if ( m_currentJob )
		delete m_currentJob;

	m_jobsList.setAutoDelete( true );
}

int TransKodeWorker::getID() const
{
	return m_workerID;
}

void TransKodeWorker::slotProfilesRemoved( const QStringList& profiles )
{
	m_jobsListMutex.lock();

	QPtrList<TransKodeJob> invalidJobs;
	invalidJobs.setAutoDelete( true );
	for ( QPtrList<TransKodeJob>::Iterator it = m_jobsList.begin(); it != m_jobsList.end(); ++it )
		if ( profiles.contains( (*it)->m_profile ) )
			invalidJobs.append( *it );

	for ( QPtrList<TransKodeJob>::Iterator it = invalidJobs.begin(); it != invalidJobs.end(); ++it )
		m_jobsList.remove( *it );

	m_jobsListMutex.unlock();


	for ( QPtrList<TransKodeJob>::Iterator it = invalidJobs.begin(); it != invalidJobs.end(); ++it )
	{
		incProfileE();
		emit jobProfileError( m_workerID, (*it)->m_jobID, (*it)->m_srcURL );
 		// config is only changed from main thread so there's no need to delay the signal
		emit mtJobProfileError( m_workerID, (*it)->m_jobID, (*it)->m_srcURL );
	}
}

void TransKodeWorker::slotProfilesModified( const QStringList& profiles )
{
	QMap<QString,ProfileOptions*> profilesOptions;
	for ( QStringList::ConstIterator it = profiles.begin(); it != profiles.end(); ++it )
		profilesOptions[*it] = m_config.profileOptions( *it );
	bool suppEncodeTypes[EncodeType::SIZE];
	for ( int idx = 0; idx < EncodeType::SIZE; ++idx )
		suppEncodeTypes[idx] = m_config.encodeTypeIsSupported( (EncodeType::TEncodeType)idx );

	m_jobsListMutex.lock();

	QPtrList<TransKodeJob> invalidJobs;
	invalidJobs.setAutoDelete( true );
	for ( QPtrList<TransKodeJob>::Iterator it = m_jobsList.begin(); it != m_jobsList.end(); ++it )
		if ( profiles.contains( (*it)->m_profile ) )
		{
			ProfileOptions* options = profilesOptions[(*it)->m_profile];
			if ( options == 0 )
				invalidJobs.append( *it );
			else if ( !options->haveValidBaseDir() || !options->haveValidNamingScheme() || !suppEncodeTypes[options->encodeType()] )
				invalidJobs.append( *it );
		}

	for ( QPtrList<TransKodeJob>::Iterator it = invalidJobs.begin(); it != invalidJobs.end(); ++it )
		m_jobsList.remove( *it );

	m_jobsListMutex.unlock();

	for ( QMap<QString,ProfileOptions*>::Iterator it = profilesOptions.begin(); it != profilesOptions.end(); ++it )
		if ( it.data() != 0 )
			delete it.data();

	for ( QPtrList<TransKodeJob>::Iterator it = invalidJobs.begin(); it != invalidJobs.end(); ++it )
	{
		incProfileE();
		emit jobProfileError( m_workerID, (*it)->m_jobID, (*it)->m_srcURL );
 		// config is only changed from main thread so there's no need to delay the signal
		emit mtJobProfileError( m_workerID, (*it)->m_jobID, (*it)->m_srcURL );
	}
}


void TransKodeWorker::slotEncTypesRemoved( const QValueList<int>& encodeTypes )
{
	QStringList allProfiles = m_config.profiles();
	QStringList invalidProfiles;
	for ( QStringList::Iterator it = allProfiles.begin(); it != allProfiles.end(); ++it )
	{
		ProfileOptions* options = m_config.profileOptions( *it );
		if ( options == 0 )
			invalidProfiles.append( *it );
		else if ( encodeTypes.contains( options->encodeType() ) )
		{
			invalidProfiles.append( *it );
			delete options;
		}
	}

	m_jobsListMutex.lock();

	QPtrList<TransKodeJob> invalidJobs;
	invalidJobs.setAutoDelete( true );
	for ( QPtrList<TransKodeJob>::Iterator it = m_jobsList.begin(); it != m_jobsList.end(); ++it )
		if ( invalidProfiles.contains( (*it)->m_profile ) )
			invalidJobs.append( *it );

	for ( QPtrList<TransKodeJob>::Iterator it = invalidJobs.begin(); it != invalidJobs.end(); ++it )
		m_jobsList.remove( *it );

	m_jobsListMutex.unlock();

	for ( QPtrList<TransKodeJob>::Iterator it = invalidJobs.begin(); it != invalidJobs.end(); ++it )
	{
		incProfileE();
		emit jobProfileError( m_workerID, (*it)->m_jobID, (*it)->m_srcURL );
 		// config is only changed from main thread so there's no need to delay the signal
		emit mtJobProfileError( m_workerID, (*it)->m_jobID, (*it)->m_srcURL );
	}
}

int TransKodeWorker::addJob( const QString& srcURL, const QString& profile )
{
	m_jobsListMutex.lock();

	int jobID = -1;

	if ( !m_threadStopped )
	{
		TransKodeJob* job = new TransKodeJob( *this, profile, srcURL );
		jobID = job->getID();
		m_jobsList.append( job );
	}

	m_jobsListMutex.unlock();

	if ( jobID != -1 )
	{
		LOG( "addJob: job added, sending wake signal to worker", WORKER );
		m_jobCondition.wakeAll();
	}

	return jobID;
}

int TransKodeWorker::addJob( const QString& srcURL, const QString& dstURL, const QString& profile, const QString tags[Tag::SIZE] )
{
	m_jobsListMutex.lock();

	int jobID = -1;

	if ( !m_threadStopped )
	{
		TransKodeJob* job = new TransKodeJob( *this, profile, srcURL, dstURL, tags );
		jobID = job->getID();
		m_jobsList.append( job );
	}

	m_jobsListMutex.unlock();

	if ( jobID != -1 )
	{
		LOG( "addJob: job added, sending wake signal to worker", WORKER );
		m_jobCondition.wakeAll();
	}

	return jobID;
}

void TransKodeWorker::addJobs( const QStringList& srcURLs, const QString& profile )
{
	if ( srcURLs.size() )
	{
		bool added = false;

		m_jobsListMutex.lock();

		if ( !m_threadStopped )
		{
			for ( QStringList::ConstIterator it = srcURLs.begin(); it != srcURLs.end(); ++it )
				m_jobsList.append( new TransKodeJob( *this, profile, *it ) );

			added = true;
		}

		m_jobsListMutex.unlock();

		if ( added )
		{
			LOG( "addJobs: jobs added, sending wake signal to worker", WORKER );
			m_jobCondition.wakeAll();
		}
	}
}

void TransKodeWorker::removeJob( int jobID )
{
	m_currentJobMutex.lock();

	QString URL;

	if ( m_currentJob )
		if ( m_currentJob->getID() == jobID )
		{
			URL = String::deepCopy( m_currentJob->m_srcURL );
			LOG( "removeJob: job in process signaled to end", WORKER );
			m_currentJob->cancel();

			// don't incJobC() here since we don't really know if the signal will
			// arrive before job ends and hence if job will be indeed cancelled
			// incJobC();

			m_currentJobMutex.unlock();

			return;
		}

	m_currentJobMutex.unlock();

	m_jobsListMutex.lock();

	bool emitSignal = false;

	for ( TransKodeJob* job = m_jobsList.first(); job != 0 ; job = m_jobsList.next() )
		if ( job->getID() == jobID )
		{
			URL = String::deepCopy( job->m_srcURL );

			LOG( "removeJob: job removed from list", WORKER );
			m_jobsList.remove();
			emitSignal = true;

			// it is correct to incJobC() here since we know for sure job was cancelled
			incJobsC();
		}

	m_jobsListMutex.unlock();

	if ( emitSignal )
	{
		emit jobCancelled( m_workerID, jobID, URL );
		// eventually a job could get removed from within the worker
		// thread so it's best to delay the signal just in case
		delaySignal( new JobCancelledEvent( jobID, URL ) );
	}
}

void TransKodeWorker::clearJobs()
{
	m_jobsListMutex.lock();

	LOG( "clearJobs: jobs removed from list", WORKER );
	m_jobsList.clear();

	m_jobsListMutex.unlock();

	m_currentJobMutex.lock();

	if ( m_currentJob )
	{
		LOG( "clearJobs: job in process signaled to end", WORKER );
		m_currentJob->cancel();
	}

	m_currentJobMutex.unlock();
}

bool TransKodeWorker::isCurrentJob( int jobID ) const
{
	m_currentJobMutex.lock();

	bool ret = m_currentJob ? (m_currentJob->getID() == jobID) : false;

	m_currentJobMutex.unlock();

	return ret;
}

int TransKodeWorker::currentJobID() const
{
	m_currentJobMutex.lock();

	int ret = m_currentJob ? m_currentJob->getID() : -1;

	m_currentJobMutex.unlock();

	return ret;
}

int TransKodeWorker::jobsQueued() const
{
	m_jobsListMutex.lock();

	int ret = m_jobsList.count();

	m_jobsListMutex.unlock();

	return ret;
}

bool TransKodeWorker::paused() const
{
	m_currentJobMutex.lock();

	bool ret = m_threadPaused != 0;

	m_currentJobMutex.unlock();

	return ret;
}

void TransKodeWorker::pause( unsigned long msecs )
{
	m_threadPausedMutex.lock();

	if ( !m_threadPaused )
	{
		LOG( "pause: pause signal sent to worker", WORKER );
		m_threadPaused = msecs;
	}

	m_threadPausedMutex.unlock();
}

void TransKodeWorker::resume()
{
	m_threadPausedMutex.lock();

	if ( m_threadPaused )
		m_threadPaused = 0;

	m_threadPausedMutex.unlock();

	LOG( "resume: resume signal sent to worker", WORKER );
	m_pauseCondition.wakeAll();
}

void TransKodeWorker::checkPaused( KProcess& process )
{
	m_threadPausedMutex.lock();

	if ( m_threadPaused )
	{
		LOG( "pollPaused: worker subprocess paused", WORKER );
		process.kill( SIGSTOP );

		LOG( "pollPaused: worker paused", WORKER );

		m_pauseCondition.wait( &m_threadPausedMutex, m_threadPaused );

		LOG( "pollPaused: worker resumed", WORKER );

		process.kill( SIGCONT );;
		LOG( "pollPaused: worker subprocess resumed", WORKER );

		m_threadPaused = 0;
	}

	m_threadPausedMutex.unlock();
}


bool TransKodeWorker::ended() const
{
	m_jobsListMutex.lock();

	bool ret = m_threadStopped;

	m_jobsListMutex.unlock();

	return ret;
}

void TransKodeWorker::end()
{
	m_jobsListMutex.lock();

	m_threadStopped = true;

	m_jobsList.clear();

	LOG( "end: remaining jobs cleared", WORKER );

	m_jobsListMutex.unlock();


	m_currentJobMutex.lock();

	if ( m_currentJob )
	{
		LOG( "end: job in process signaled to end", WORKER );
		m_currentJob->cancel();
	}

	m_currentJobMutex.unlock();

	LOG( "end: wake signal sent to worker", WORKER );
	m_jobCondition.wakeAll(); // wake the worker thread (if it was sleeping) so it ends normally
}

TransKodeJob* TransKodeWorker::getNextJob()
{
	m_jobsListMutex.lock();

	TransKodeJob* job = m_jobsList.take( 0 );

	m_jobsListMutex.unlock();


	m_currentJobMutex.lock();

	if ( m_currentJob )
		delete m_currentJob;
	m_currentJob = job;

	m_currentJobMutex.unlock();

	return job;
}

void TransKodeWorker::incProfileE() { m_statsMutex.lock(); m_profileE++; m_statsMutex.unlock();}
void TransKodeWorker::resetProfileE() { m_statsMutex.lock(); m_profileE = 0; m_statsMutex.unlock();}
int TransKodeWorker::profileE() { m_statsMutex.lock(); int ret = m_profileE; m_statsMutex.unlock(); return ret; }

void TransKodeWorker::incJobsC() { m_statsMutex.lock(); m_jobsC++; m_statsMutex.unlock();}
void TransKodeWorker::resetJobsC() { m_statsMutex.lock(); m_jobsC = 0; m_statsMutex.unlock();}
int TransKodeWorker::jobsC() { m_statsMutex.lock(); int ret = m_jobsC; m_statsMutex.unlock(); return ret; }

void TransKodeWorker::run()
{
	LOG( "run: worker started", WORKER );

	m_jobsListMutex.lock();

	while ( !m_threadStopped ) // NOTE: access to m_threadStopped is guarded by m_jobsĹistMutex
	{
		LOG( "run: worker going to sleep", WORKER );

		// Will unlock the given mutex and then wait. After it's
		// done waiting, it will lock the mutex again and return.
		m_jobCondition.wait( &m_jobsListMutex );
		m_jobsListMutex.unlock();

		LOG( "run: worker awaken", WORKER );

		resetProfileE();
		resetJobsC();

		m_streamRipE = 0;
		m_streamRipF = 0;

		m_transcodeE = 0;
		m_tSrcFiles.clear();
		m_tDstFiles.clear();

		m_copyE = 0;
		m_cSrcFiles.clear();
		m_cDstFiles.clear();

		TransKodeJob* job;
		while ( ( job = getNextJob() ) )
			job->transcode();

		int pE = profileE();
		int jC = jobsC();

		emit roundFinished( m_workerID );
		delaySignal( new RoundFinishedEvent() );

		emit roundStats(
			m_workerID,
			m_streamRipE, m_streamRipF,
			m_transcodeE, m_tSrcFiles.size(),
			m_copyE, m_cSrcFiles.size(),
			jC, pE );
		delaySignal( new RoundShortStatsEvent(
			m_streamRipE, m_streamRipF,
			m_transcodeE, m_tSrcFiles.size(),
			m_copyE, m_cSrcFiles.size(),
			jC, pE ) );

		emit roundStats( m_workerID,
			m_streamRipE, m_streamRipF,
			m_transcodeE, m_tSrcFiles, m_tDstFiles,
			m_copyE, m_cSrcFiles, m_cDstFiles,
			jC, pE );
		delaySignal( new RoundLongStatsEvent(
			m_streamRipE, m_streamRipF,
			m_transcodeE, m_tSrcFiles, m_tDstFiles,
			m_copyE, m_cSrcFiles, m_cDstFiles,
			jC, pE ) );

		m_jobsListMutex.lock();
	}

	m_jobsListMutex.unlock();

	LOG( "run: worker finished", WORKER );
}

void TransKodeWorker::delaySignal( TransKodeWorkerEvent* event )
{
	kapp->postEvent( this, event );
}

void TransKodeWorker::customEvent( QCustomEvent* e )
{
	switch ( e->type() )
	{
		case JOBPROFILEERROREVENT:
			emit mtJobProfileError( m_workerID, ((JobProfileErrorEvent*)e)->m_jobID, ((JobProfileErrorEvent*)e)->m_URL );
			break;
		case JOBCANCELLEDEVENT:
			emit mtJobCancelled( m_workerID, ((JobCancelledEvent*)e)->m_jobID, ((JobCancelledEvent*)e)->m_URL );
			break;
		case STREAMRIPSTARTEDEVENT:
			emit mtStreamRipStarted( m_workerID, ((StreamRipStartedEvent*)e)->m_jobID, ((StreamRipStartedEvent*)e)->m_URL );
			break;
		case STREAMRIPFINISHEDEVENT:
			emit mtStreamRipFinished( m_workerID, ((StreamRipFinishedEvent*)e)->m_jobID, ((StreamRipFinishedEvent*)e)->m_URL, ((StreamRipFinishedEvent*)e)->m_rippedFiles );
			break;
		case STREAMRIPERROREVENT:
			emit mtStreamRipError( m_workerID, ((StreamRipErrorEvent*)e)->m_jobID, ((StreamRipErrorEvent*)e)->m_URL, ((StreamRipErrorEvent*)e)->m_errorMsg );
			break;
		case COPYSTARTEDEVENT:
			emit mtCopyStarted( m_workerID, ((CopyStartedEvent*)e)->m_jobID, ((CopyStartedEvent*)e)->m_srcFile, ((CopyStartedEvent*)e)->m_dstFile );
			break;
		case COPYFINISHEDEVENT:
			emit mtCopyFinished( m_workerID, ((CopyFinishedEvent*)e)->m_jobID, ((CopyFinishedEvent*)e)->m_srcFile, ((CopyFinishedEvent*)e)->m_dstFile );
			break;
		case COPYERROREVENT:
			emit mtCopyError( m_workerID, ((CopyErrorEvent*)e)->m_jobID, ((CopyErrorEvent*)e)->m_srcFile, ((CopyErrorEvent*)e)->m_errorMsg );
			break;
		case TRANSCODESTARTEDEVENT:
			emit mtTranscodeStarted( m_workerID, ((TranscodeStartedEvent*)e)->m_jobID, ((TranscodeStartedEvent*)e)->m_srcFile, ((TranscodeStartedEvent*)e)->m_dstFile );
			break;
		case TRANSCODEFINISHEDEVENT:
			emit mtTranscodeFinished( m_workerID, ((TranscodeFinishedEvent*)e)->m_jobID, ((TranscodeFinishedEvent*)e)->m_srcFile, ((TranscodeFinishedEvent*)e)->m_dstFile );
			break;
		case TRANSCODEERROREVENT:
			emit mtTranscodeError( m_workerID, ((TranscodeErrorEvent*)e)->m_jobID, ((TranscodeErrorEvent*)e)->m_srcFile, ((TranscodeErrorEvent*)e)->m_errorMsg );
			break;
		case TRANSCODEPROGRESSEVENT:
			emit mtTranscodeProgress( m_workerID, ((TranscodeProgressEvent*)e)->m_jobID, ((TranscodeProgressEvent*)e)->m_progress );
			break;
		case ROUNDFINISHEDEVENT:
			emit mtRoundFinished( m_workerID );
			break;
		case ROUNDSHORTSTATSEVENT:
			emit mtRoundStats(
				m_workerID,
				((RoundShortStatsEvent*)e)->m_streamRipE, ((RoundShortStatsEvent*)e)->m_streamRipF,
				((RoundShortStatsEvent*)e)->m_transcodeE, ((RoundShortStatsEvent*)e)->m_transcodeF,
				((RoundShortStatsEvent*)e)->m_copyE, ((RoundShortStatsEvent*)e)->m_copyF,
				((RoundShortStatsEvent*)e)->m_jobsC, ((RoundShortStatsEvent*)e)->m_profileE );
			break;
		case ROUNDLONGSTATSEVENT:
			emit mtRoundStats(
				m_workerID,
				((RoundLongStatsEvent*)e)->m_streamRipE, ((RoundLongStatsEvent*)e)->m_streamRipF,
				((RoundLongStatsEvent*)e)->m_transcodeE, ((RoundLongStatsEvent*)e)->m_tSrcFiles, ((RoundLongStatsEvent*)e)->m_tDstFiles,
				((RoundLongStatsEvent*)e)->m_copyE, ((RoundLongStatsEvent*)e)->m_cSrcFiles, ((RoundLongStatsEvent*)e)->m_cDstFiles,
				((RoundLongStatsEvent*)e)->m_jobsC, ((RoundLongStatsEvent*)e)->m_profileE );
			break;
		default:
			QObject::event( e );
			break;
	}
}
