#ifndef _TRANSKODEFILEJOB_H_
#define _TRANSKODEFILEJOB_H_

/***************************************************************************
 *   Copyright (C) 2005 by Sergio Pistone                                  *
 *   sergio_pistone@yahoo.com.ar                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <commondefs.h>
#include <transkodeconfig.h>

#include <kprocess.h>
#include <qmutex.h>

class TransKodeFileWorker;
class TransKodeFileTypeResolver;

class TransKodeFileJob : public QObject
{
	Q_OBJECT

	friend class TransKodeFileWorker;

	public:
		typedef enum { trans = 0, copy = 1 } JobType;

		TransKodeFileJob( const TransKodeConfig& config, const QString& srcPath, const QString& profile );
		TransKodeFileJob( const TransKodeConfig& config, const QString& srcPath );
		virtual ~TransKodeFileJob();

		int id() const;

		bool enqueued() const;
		int queuePos() const;
		int progress() const;
		bool inProgress() const;
		bool profileError() const;

		QString profile() const;
		bool setProfile( const QString& profile );

		QString srcPath() const;
		QString dstPath() const;

		QString* tags() const;
		bool setTags( const QString* tags );

		bool refresh();

		TransKodeFileWorker* worker() const;

		int enqueue( TransKodeFileWorker* worker );
		bool tryCancel();


	signals:
		/**
		 * Warning! this signals MAY NOT be emitted from the main thread (see qt docs on threads)
		 */
		void profileErrorSet( int jobID, const QString& srcPath );
		void profileErrorUnset( int jobID, const QString& srcPath );

		void dstPathChanged( int jobID, const QString& dstPath );

		void enqueued( int jobID, int queuePos );
		void posChanged( int jobID, int queuePos );
		void cancelled( int jobID );

		// processing state
		void started( int jobID, const QString& srcPath, const QString& dstPath, int type );
		void finished( int jobID, const QString& srcPath, const QString& dstPath, int type );
		void error( int jobID, const QString& srcPath, const QString& errorMsg, int type );
		void progress( int jobID, int progress, int type );

		/**
		 * This signals ARE emitted from the main thread
		 */
		void mtProfileErrorSet( int jobID, const QString& srcPath );
		void mtProfileErrorUnset( int jobID, const QString& srcPath );

		void mtDstPathChanged( int jobID, const QString& dstPath );

		void mtEnqueued( int jobID, int queuePos );
		void mtPosChanged( int jobID, int queuePos );
		void mtCancelled( int jobID );

		// processing state
		void mtStarted( int jobID, const QString& srcPath, const QString& dstPath, int type );
		void mtFinished( int jobID, const QString& srcPath, const QString& dstPath, int type );
		void mtError( int jobID, const QString& srcPath, const QString& errorMsg, int type );
		void mtProgress( int jobID, int progress, int type );

	protected:

		// called by the worker when the job position in the queue changes
		void advanceQueuePos();

		QString generateDstPath();
		void setDstPath( const QString& dstPath );

		QString resolveTagForFile( Tag::TTag tag, bool quote=false );
		QString resolveTagForPath( Tag::TTag tag, bool quote=false );

		void readTags();
		void writeTags( EncodeType::TEncodeType encodeType );

		bool monitorProcess( KProcess& process );

		void transcode();
		int decodeToWav( const QString& srcPath, const QString& dstPath );
		int encodeFromWav( const QString& srcPath, const QString& dstPath );

		bool cancelRequested() const;
		bool cancelled() const;

		void setProgress( int progress, int type );
		void normalExit( int type );
		void abnormalExit( const QString& errorMsg, int type );

		void delaySignal( QCustomEvent* event );
		virtual void customEvent( QCustomEvent* );

	protected:
		const int m_jobID;

		const TransKodeConfig& m_config;

		mutable QMutex m_jobMutex;

		TransKodeFileWorker* m_worker;
		int m_queuePos;		// -1: not queued / cancelled
							//  0: transcoding
							// >0: queued

		int m_cancelled;	//  0: no cancel requested
							//  1: cancel requested
							//  2: cancel request accepted

		int m_progress;		// -1: not started
							//  0: started
							// >=0 && <100: in progress

		QString m_profile;
		ProfileOptions* m_options;

		QString m_srcPath;
		QString m_dstPath;
		QString m_fileTags[Tag::SIZE];
		QString m_userTags[Tag::SIZE];

		static TransKodeFileTypeResolver s_fileTypeResolver;
		static QMutex s_fileTypeResolverMutex;
};

#endif
